
# coding: utf-8

# In[1]:


# %matplotlib inline
import numpy as np
from random import randint,uniform,choice,gauss,sample
import matplotlib.pyplot as plt
from scipy.spatial import distance
from scipy.stats.stats import pearsonr
from  scipy.stats import truncnorm
import matplotlib.cm as cm
from math import exp,sin,cos,pi,sqrt
import pickle
from copy import deepcopy
import pandas as pd
from mpl_toolkits.mplot3d import Axes3D
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from copy import copy


def run_model(N=500,S=3,T=1000,initial_parameters = None,e = 0,zeta = 0,neighborhood = 'mean_field',
              perception_geometry='balance',movement_geometry='balance_equilibrium',
              noise_geometry='euclidean', threshold=0,plot_interval = 10,
              continuous_plotting=False,continuous_metrics=False,pol_met='cityblock',pol_exp=2, location='',
              sigmoid=True,alpha=.4,gamma=.5,suffix='',check_interval=10,convergence_thresh=0,
              return_matrices=False,weighted_average=False,reweight_function=False,funct=None):
    """This is the main function to run the ABM. Parameters:
    N: Number of agents
    S: Number of opinion dimensions
    T: Maximum number of time steps the model iterates over
    initial_parameters: If None, distribute opinion vectors uniformly at random over opinion space in initial state.
    neighborhood: Defines which agents can interact; 'mean_field' means every agent can interact with every other agent.
    perception_geometry: How agents translate their relative opinion vectors into similarity/ relation; 'balance' is the mechanism described in our paper.
    movement_geometry: How agents exchange opinions; 'balance_equilibrium' is the modus described in our paper.
    zeta: The noise level in the model.
    threshold: An optional interaction threshold between 0 and 1, for simulating bounded confidence.
    continuous_plotting: Whether or not to continuously create scatter plots and similarity histograms. 
    plot_interval: Time steps between consecutive plots
    continuous_metrics: Whether or not to continuously compute metrics (polarization, opinion constraint etc.).
    pol_met: Distance to use in polarization metric; default is cityblock (manhattan distance).
    pol_exp: Exponent to use in polarization metric; default is 2.
    location: Where to save plots
    sigmoid: Whether to use sigmoid function (resembling evaluative extremeness) or not (for positivity bias).
    alpha: time scale parameter, by how much agents approach the perfectly balanced vector in each interaction.
    gamma: general bias of the evaluative extremeness function (not used in paper)
    e: evaluative extremeness parameter; default is 0 (no evaluative extremeness)
    suffix: suffix of plots
    check_interval: For how many time steps to check for convergence
    convergence_thresh: Threshold for model convergence (average difference between consecutive opinion vectors)
    return_matrices: whether to return matrices (opinion matrix etc) in model object.
    weighted_average: whether to use weighted average (not described in paper) for aggregating SGMs from different opinion dimensions
    reweight_function: Whether to use the reweighting function taken from parameter 'funct'.

    run_model returns a dictionary object containing the initial parameters, and metrics of the last time step of the model, such as polarization, constraint, extremeness etc.
    """
    epsilon =  1- e
    del e

    # Counter:
    c = 0
    # List of polarization values:
    pol_list = []
    # Create initial opinion matrix:
    opinion_mat = initiate_opinion_mat(N,S,initial_parameters)
    old_opinion_mat = copy(opinion_mat)
    # Create emotion matrix:
    epsilon_mat = initiate_emo_mat(N=N,initial_parameters=epsilon,max_val=2)
    # Create gamma matrix:
    gamma_mat = initiate_emo_mat(N=N,initial_parameters=gamma)    
    # Similarity vector:
    sim_vec = []
    # Standard Deviation vector:
    sd_vec = []
    # Variance explained by first component vector:
    pc_vec = []
    # Average correlation:
    av_corr_vec = []
    # Difference Vector:
    diff_vec = []
    # Extremeness vector:
    extreme_vec = []
    #Mono_polarization vector:
    mono_pol_vec = []
    # Iterate over time steps:
    for t in range(T):
        rand_i_vec = sample(range(N),N)
        if (t % plot_interval == 0) or t == (T-1):
            if continuous_plotting == True:
                #Create scatter plot:
                scatter3d_plot(opinion_mat,loc= location + 'ScatterFrame_' + str(c) + '.png')
                if c > 0:
                    sim_hist(sim_vec,loc= location + 'SimHist_' + str(c) + '.png')
                    sim_vec = []
                #Compute polarization,average standard deviation, extremeness, 
                #average correlation, and mono-polarization:
            if continuous_metrics == True:
                pol_list.append(polarization(opinion_mat,distmet=pol_met,ex=pol_exp))
                sd_vec.append(sum(np.std(opinion_mat,axis=0))/S)
                #pc_vec.append(first_comp(opinion_mat))
                extreme_vec.append(abs(opinion_mat).sum()/(N*S))
                av_corr_vec.append(average_corr(opinion_mat))
                mono_pol_vec.append(mono_pol(opinion_mat))
            c += 1  
        for n in range(N):
            # Every plot_interval timesteps, make a scatter plot:             
            # Select random pair of agents i and j:
            i,j = select_agents(N,n,rand_i_vec,neighborhood,neighbor_mat=None)
            # Determine similarity between opinion vectors of i and j:
            sim = similarity(opinion_mat[i],opinion_mat[j],perception_geometry=perception_geometry,
                             weighted_average=weighted_average)
            sim_vec.append(sim)
            # Reweight similarty:
            sim = reweight(sim,exp=epsilon_mat[i],ref_point=gamma_mat[i],sigmoid=sigmoid,reweight_function=reweight_function,funct=funct)
            # Interaction threshold:
            if not check_threshold(sim,thresh=threshold):
                 continue
            # Exchange opinions between i and j (j influences i):
            opinion_mat[i] = exchange(opinion_mat[i],opinion_mat[j],sim,movement_geometry,
                                      alpha=alpha)
            # Induce noise:
            rand_n = randint(0,N-1)
            opinion_mat[rand_n] = add_noise(opinion_mat[rand_n],zeta,noise_geometry)
        # Compute sum of absolute differences of old matrix and new one:
        if (t % check_interval == 0):
            # print(t)
            diff = abs(opinion_mat - old_opinion_mat).sum()/(N*S*2)
            diff_vec.append(diff)
            old_opinion_mat = copy(opinion_mat)
            if (len(diff_vec) > 4) and np.all(np.array(diff_vec[-5:]) <= convergence_thresh):
                break
    if pol_met:
        outcome_dict = {}
        outcome_dict['Method'] = movement_geometry
        outcome_dict['E'] = epsilon
        outcome_dict['A'] = alpha
        outcome_dict['Z'] = zeta
        outcome_dict['T'] = t + 1
        outcome_dict['D'] = S
        outcome_dict['G'] = gamma
        # Polarization metric:
        outcome_dict['pol'] = polarization(opinion_mat,distmet=pol_met,ex=pol_exp)
        # Average standard deviation:
        outcome_dict['sd'] = sum(np.std(opinion_mat,axis=0))/S
        outcome_dict['pc'] = first_comp(opinion_mat)
        outcome_dict['extreme'] = abs(opinion_mat).sum()/(N*S)
        # Average correlation (opinion constraint):
        outcome_dict['corr'] = average_corr(opinion_mat)
        outcome_dict['mono_pol'] = mono_pol(opinion_mat)
        if continuous_metrics == True:
            outcome_dict['pol_vec'] = pol_list
            outcome_dict['sd_vec'] = sd_vec
            outcome_dict['pca_vec'] = pc_vec
            outcome_dict['av_cor_vec'] = av_corr_vec
            outcome_dict['extreme_vec'] = extreme_vec
            outcome_dict['mono_pol_vec'] = mono_pol_vec
        if return_matrices == True:
            outcome_dict['mat'] = opinion_mat
            outcome_dict['epsilon_mat'] = epsilon_mat
            outcome_dict['gamma_mat'] = gamma_mat
            # outcome_dict['difference_vector'] = diff_vec
        #save_loc = '_A' + str(alpha) + '_E' + str(reweight_exp) +'_Z' + str (zeta) + "_" + suffix +'.png'
        #scatter3d_plot(opinion_mat,loc = location + 'ScatterFrame3D' + save_loc)
        #scatter_plot(opinion_mat,loc = location + 'ScatterFrame' + save_loc)
        #line_plot(pol_vec=pol_list,sd_vec=sd_vec,pc_vec=pc_vec,plot_interval=plot_interval,loc= location + 'PolarizationTS' + save_loc)                
    return(outcome_dict)

# In[2]:


# Utility Functions:
def versor(v):
    """Computes versor (unit vector with same direction) of v."""
    return(v / vlen(v))

def vlen(v):
    """Returns length of vector v."""
    return(np.linalg.norm(v))

def vrescale(v,l):
    """Rescales vector v to length l."""
    return(np.asarray(v)*(l/vlen(v)))

def trunc_vec(S=1,sigma=.1,mu=0,min_val=-1,max_val=1):
    """Create a vector of length S with the components drawn from a truncated normal 
    distribution between [-1,1], with standard deviation sigma.""" 
    a, b = (min_val - mu) / sigma, (max_val - mu) / sigma
    return(truncnorm(a, b, loc=mu, scale=sigma).rvs(S))

def randnormal_rescale(v,mu=0,sigma=.1):
    """ Returns the unit vector of the vector,multiplied by a random normal number."""
    if sigma != 0:
        return(vrescale(v, truncnorm((0 - mu) / sigma, (1 - mu) / sigma, loc=mu, scale=sigma).rvs()))
    else:
        return(vrescale(v,mu))

def rotate(v1,v2,r,scaling='original_length'):
    """Rotates v1 in the direction of v2 by r, which should be from [-1,1]. 'scaling'
    defines the length of the output vector. Default is 'original_length'."""
    v1 = np.array(v1)
    v2 = np.array(v2)
    if scaling == 'original_length':
        new_length = vlen(v1)
    elif scaling == 'mean_length':
        new_length = (1 - abs(r))*vlen(v1) + r*vlen(v2)
    #if r is negative, take v2 * - 1, and make r positive:
    if r == 0:
        return(versor(v1)*(new_length))
    v2 = np.sign(r)*v2
    r = abs(r)    
    # Compute angles:
    alpha = angle_between(v1, v2,normalize=False)
    theta = alpha*r
    # Compute p (weighting parameter):
    p = (np.sin(theta)*np.cos(alpha/2))/(np.cos((alpha/2)-theta)*np.sin(alpha))
    # compute weighted mean of v1 and v2:
    mean_v = (1 - p)*versor(v1) + p*versor(v2)
    # Bring rotated vector back to demanded size:
    return(mean_v*(new_length/vlen(mean_v)))

def angle_between(v1, v2,normalize=True):
    " Returns the smaller angle in radians between vectors 'v1' and 'v2' in pi radians."
    v1_u = versor(v1)
    v2_u = versor(v2)
    if normalize:
        return(np.arccos(np.clip(np.dot(v1_u, v2_u.T), -1.0, 1.0))/pi)
    else:
        return(np.arccos(np.clip(np.dot(v1_u, v2_u.T), -1.0, 1.0)))

def circle_point(x,r=1,sign = 1):
    """Computes y-coordinate for x on circle of radius r."""
    return(sign*sqrt(r**2 - x**2))

def euc_approach(v1,v2,w,alpha=.5):
    """Returns weighted mean of v1 and v2 (r is weight of v2)."""
    return(v1 - alpha*w*(v1-v2))


def sim_to_rel(sim,ref_point=.5):
    """Transforms similarity ([0,1]) to relation ([-1,1]) values. 
    Resets the reference point (where sim_to_rel(sim) = 0) to ref_point."""
    if sim >= ref_point:
        return(rescale(sim,(ref_point,1),(0,1)))
    elif sim < ref_point:
        return(rescale(sim,(0,ref_point),(-1,0)))


def rescale(x, range1, range2):
    """rescales a value x within range1 to a value within range2."""
    delta1 = range1[1] - range1[0]
    delta2 = range2[1] - range2[0]
    return(delta2 * (x - range1[0]) / delta1) + range2[0]
    
def rel_reweight(R,exp=1,sigmoid=True):
    """Reweigh the relation R by exponent exp (either sigmoid or not),
    and return a similarity value."""
    if exp == 1:
        return((R+1)/2)
    elif sigmoid:
        R = (abs(R)**exp)*np.sign(R)
        return((R+1)/2)
    else:
        sim = (R+1)/2
        sim = sim**exp
        return(sim)
    
def reweight(sim,exp=1,ref_point=.5,sigmoid=True,reweight_function=False,funct=None):
    if reweight_function == True:
        reweighted = funct((sim-.5)*2)
        return((reweighted+1)/2)
    return(rel_reweight(sim_to_rel(sim,ref_point=ref_point),exp=exp,sigmoid=sigmoid))

    
# Signed geometric mean:
def sgm(v1,v2):
    "Signed geometric mean of v1 and v2."
    m = v1*v2
    return(np.sign(m)*(abs(m)**(1./2.)))


# Relation between v1 and v2: 
def relation(v1,v2,weighted_average=False):
    """Relation between v1 and v2. Computes the signed geometric mean 
    of v1 and v2. If weighted_average == True, computes the average of SGM(v1,v2),
    weighted by abs(v1). Otherwise, computes the arithmetic mean of the SGM of v1 and v2."""
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    # Using the average of the SGM, weighted by the opinion wights of v1:
    if weighted_average== True:
        # In case all components of v1 are zero,return zero:
        if not np.count_nonzero(v1):
            return(0)#np.average(sgm(v1, v2))
        # Otherwise, return the weighted average:
        else:
            return(np.squeeze(np.average(sgm(v1, v2),weights=abs(v1))))
    # Using the average (non-weighted) of the SGM:
    else:
        return(sum(sgm(v1, v2))/len(v1))
    
# Balance vector between v1 and v2, given relation R:
def balance(v1,v2,R):
    """Balance vector between v1 and v2, given relation R. If geometric=True, computes the
    signed geometric mean between v1, v2 and R."""
    m = v1*v2*R
    return(np.sign(m)*(abs(m)**(1./3.)))

# Change in vector v given balance vector B. Change is either proportional to balance, or 
# proportional to 1 - Balance. Parameter alpha determines magnitude of change:
def balance_change(v,B,alpha=.5,proportional=False):
    """Change in vector v given balance vector B. Change is either proportional to balance, or 
    proportional to 1 - Balance. Parameter alpha determines magnitude of change."""
    if proportional:
        # Change magnitude proportional to Balance:
        # (Vector + Balance times Sign of Vector times 1 - magnitude times beta):
        return(v+(B*alpha*np.sign(v)*(1-abs(v))))
    else:
        # Change magnitude proportional to 1 - Balance:
        # (Vector + (1 minus Balance)/2 times Sign of Vector times 1 - magnitude times beta:
        return(v+((1-B)*alpha*np.sign(B)*np.sign(v)*(1-abs(v))))
    
def balance_exchange(v1,v2,alpha=.5,proportional=False,weighted_average=False):
    """Opinion exchange between v1 and v2 according to continuous balance theory."""
    return(balance_change(v1,balance(v1,v2,relation(v1,v2,weighted_average=weighted_average)),alpha=alpha, proportional=proportional))

def balanced_vec(v1,v2,R,alpha=.5,length_weighted=False):
    """Return a mix between v1 and the balanced vector."""
#     R = relation(v1,v2,geometric=True)
#     S = (R+1)/2
#     S = reweight(S,exp=e)
#     R = (S-0.5)*2
    m = v2*R
    m = np.sign(m)*(abs(m)**(1./2.))
    if length_weighted:
        return((v1*(abs(v1))) + (m*(1-abs(v1))))
    else:
        return((v1*(1-alpha)) + (m*alpha))

def wiest(v1,v2):
    """Prediction of a third relation in a triangle according to Wiest's (1965) 
    quantitative extension of Balance Theory, as presented in Wellens and Thistlethwaite (1971)."""
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    out_vec = np.zeros(len(v1))
    for i in range(len(v1)):
        Zu = 1 - abs(v1[i] - v2[i])
        Zl = abs(v1[i] + v2[i]) - 1
        out_vec[i] = (Zu + Zl)/2
    return(out_vec)

def feather(v1,v2):
    """Prediction of a third relation in a triangle according to Feather's (1964) 
    Structural Balance Model of Communication Effects, as presented in Wellens and Thistlethwaite (1971)."""
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    out_vec = np.zeros(len(v1))
    for i in range(len(v1)):
        if (v1[i] != 0) and (v2[i] != 0):
            if np.sign(v1[i]) == np.sign(v2[i]):
                out_vec[i] = 1 - abs(v1[i] - v2[i])
            else:
                out_vec[i] = - (abs(v1[i] - v2[i])/2)
        else:
            out_vec[i] = ((1 - abs(v1[i] - v2[i])) + (- (abs(v1[i] - v2[i])/2)))/2
    return(out_vec)


def wiest_move(v1,v2,R,alpha=.4):
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    R = np.array([R]*len(v2))
    balanced_vec = wiest(v2,R)
    new_vec = (v1*(1-alpha)) + (balanced_vec*alpha)
    return(new_vec)

def feather_move(v1,v2,R,alpha=.4):
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    R = np.array([R]*len(v2))
    balanced_vec = feather(v2,R)
    # print(balanced_vec)
    new_vec = (v1*(1-alpha)) + (balanced_vec*alpha)
    return(new_vec)


def first_comp(m):
    """Performas a PCA on matrix m and returns the fraction of variance explained by the first component."""
    sm = StandardScaler().fit_transform(m)
    pca = PCA(n_components=1)
    principalComponents = pca.fit_transform(sm)
    return(pca.explained_variance_ratio_[0])


# In[3]:


# Opinion exchange:
def exchange(v1,v2,sim,movement_geometry='hypersphere_repulsion',alpha=.5):
    # Vector rotation towards (sim > .5) or away from (sim < .5) the other opinion vector:
    if movement_geometry == 'hypersphere_repulsion':
        return(rotate(v1,v2,(2*sim)-1))
    
    # As above, with change of vector length:
    if movement_geometry == 'hypersphere_repulsion_rescale':
        return(rotate(v1,v2,(2*sim)-1,scaling='mean_length'))
    
    # Vector rotation towards other vector, proportional to sim: 
    elif movement_geometry == 'hypersphere_attraction':
        return(rotate(v1,v2,sim))
    
    # Euclidean approach (sim > .5) or repulsion (sim < .5) from other vector:
    elif movement_geometry == 'euclidean_repulsion':
        return(np.clip(euc_approach(v1,v2,(2*sim)-1,alpha=alpha),-1,1))
    
    # Euclidean approach proportional to sim:
    elif movement_geometry == 'euclidean_attraction':
        return(euc_approach(v1,v2,sim))
    
    # Mean of two vectors:
    elif movement_geometry == 'euclidean_mean':
        return(euc_approach(v1,v2,0.5))
    
    # Rotation and growth/ shrinking:
    elif movement_geometry == 'hypersphere_balance':
        v = rotate(v1,v2,(2*sim)-1)
        vl = vlen(v1)
        growth = abs((2*sim)-1) - .5
        # Growing:
        if growth > 0:
            # grow rotated vector by 1 - vector length times growth factor:
            return(vrescale(v,vl + alpha*((1-vl)*growth)))
        # Shrinking:
        elif growth < 0:
            # shrink rotated vector by vector length times shrink factor:
            return(vrescale(v,vl + alpha*((vl)*growth)))
        #Neither:
        else:
            return(v)
            
    elif movement_geometry == 'dot_clip':
        growth = (2*sim) - 1
        return(np.clip((v1 + 0.05*(growth*v2)),-1,1))
        
    # 
    elif movement_geometry == 'dissonance_reduction':
        d = dissonance(v1,v2)
        d = np.clip(d, -1.0, 0)
        return(v1 + (d*v1))
    
    # Change according to balance in opinion dimension:
    elif movement_geometry == 'balance_proportional':
        R = (sim - 0.5)*2
        B = balance(v1,v2,R)
        return(balance_change(v1,B,alpha=alpha,proportional=True))
    
    # Change according to inverse of balance in opinion dimension: 
    elif movement_geometry == 'balance_inverse':
        R = (sim - 0.5)*2
        B = balance(v1,v2,R)
        return(balance_change(v1,B,alpha=alpha,proportional=False))
    
    # Compute the balance vector of v1 and v2 given their similarity, then move v1 by alpha:
    elif movement_geometry == 'balance_equilibrium':
        R = (sim - 0.5)*2
        return(np.clip(balanced_vec(v1,v2,R,alpha=alpha),-1,1))

    # Movement according to Feather:
    elif movement_geometry == 'feather':
        R = (sim - 0.5)*2
        return(np.clip(feather_move(v1,v2,R,alpha=alpha),-1,1))

    # Movement according to Wiest:
    elif movement_geometry == 'wiest':
        R = (sim - 0.5)*2
        return(np.clip(wiest_move(v1,v2,R,alpha=alpha),-1,1))
    
    # After Flache and Maes 2008 ("Why do faultlines matter?")
    elif movement_geometry == 'cityblock_repulsion':
        w = (2*sim) - 1
        delta = w*(v2-v1)
        for c in range(len(delta)):
            if delta[c] > 0:
                delta[c] = delta[c]*(1-v1[c])
            else:
                delta[c] = delta[c]*(1+v1[c])
        return(v1 + delta)


# In[4]:


# Random agent selection: 
def select_agents(N,n,rand_i_vec,neighborhood='mean_field',neighbor_mat=None):
    i = rand_i_vec[n]
    j = i
    while j == i:
        j = randint(0,N-1)
    return((i,j))


# In[5]:


#### Perceivced opinion similarity between agents:
def similarity(v1,v2,perception_geometry='hypersphere',weighted_average=False):
    # Angle between vectors, normalized between 0 (180 degrees) and 1 (0 degrees):
    if perception_geometry == 'hypersphere':
        return(1 - np.arccos(np.clip(np.dot(versor(v1),versor(v2)),-1,1)) /pi)
    # Euclidean distance, normalized by maximum possible distance:
    elif perception_geometry == 'euclidean':
        return(1 - (distance.euclidean(v1,v2))/sqrt(4*len(v1)))
    # Block distance:
    elif perception_geometry == 'cityblock':
        return(1- (distance.cityblock(v1,v2)/(2*len(v1))))
    # Dot product
    elif perception_geometry == 'dot':
        return((np.clip(((np.dot(v1, v2))), -1, 1) + 1)/2)
    # Balance Relation, rescaled between 0 and 1:
    elif perception_geometry == 'balance':
        R = relation(v1,v2,weighted_average=weighted_average)
        return((R+1)/2)
    elif perception_geometry == 'feather':
        R = sum(feather(v1, v2))/len(v1)
        return((R+1)/2)
    elif perception_geometry == 'wiest':
        R = sum(wiest(v1, v2))/len(v1)
        return((R+1)/2)
    


# In[6]:


### Plotting functions:
# Create a Scatter Plot of the opinion vectors (first and second dimension):
def scatter_plot(opinion_mat,save_f=True,loc='Scatter.png'):
    plt.xlim([-1.1,1.1])
    plt.ylim([-1.1,1.1])
    plt.xticks(np.linspace(-1,1,5))
    plt.yticks(np.linspace(-1,1,5))
    plt.axes().set_aspect('equal')
    plt.grid()
    plt.scatter(opinion_mat[:,0],opinion_mat[:,1],s=5,alpha=.5)
    if save_f == True:
        plt.savefig(loc,bbox_inches='tight',dpi=700)
        plt.close()
    else:
        plt.show()
    
def scatter3d_plot(opinion_mat,save_f=True,loc='Scatter_3D.pdf'):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.set_xlim(-1,1)
    ax.set_ylim(-1,1)
    ax.set_zlim(-1,1)
    ax.set_xticks(np.linspace(-1,1,5))
    ax.set_yticks(np.linspace(-1,1,5))
    ax.set_zticks(np.linspace(-1,1,5))
    ax.tick_params(axis='both', which='major', labelsize=12)
    ax.set_xlabel(r'$d_1$',fontsize=14)
    ax.set_ylabel(r'$d_2$',fontsize=14)
    ax.set_zlabel(r'$d_3$',fontsize=14)
    #ax.axes().set_aspect('equal')
    ax.scatter(opinion_mat[:,0], opinion_mat[:,1], opinion_mat[:,2])
    if save_f == True:
        fig.savefig(loc,bbox_inches='tight',pad_inches=0.185,dpi=700,transparent=True)
        plt.close()
    else:
        fig.show()
    
def quiver_plot(opinion_mat,loc):
    plt.xlim([-1.1,1.1])
    plt.ylim([-1.1,1.1])
    plt.axes().set_aspect('equal')
    plt.quiver(0,0,opinion_mat[:,0],opinion_mat[:,1],s=2,alpha=.9)
    plt.savefig(loc,bbox_inches='tight',dpi=700)
    plt.close()
    
def line_plot(pol_vec,sd_vec,pc_vec,plot_interval,loc='PolarizationTS.png'):
    fig = plt.figure()
    ax = fig.add_subplot(1,1,1) 
    ax.set_ylim([0,1])
    ax.set_xlabel('Timestep')
    #ax.ylabel('Polarization')
    x_vec = [x*plot_interval for x in range(len(pol_vec))]
    ax.plot(x_vec,pol_vec,c='r',label='Polarization')
    ax.plot(x_vec,sd_vec,c='b',label='Standard Dev.')
    ax.plot(x_vec,pc_vec,c='g',label='PCA Var. expl.')
    ax.grid()
    ax.legend()
    fig.savefig(loc,bbox_inches='tight',dpi=700)
    plt.close(fig)
    
def sim_hist(sim_vec,loc):
    plt.xlim([0,1])
    plt.ylim([0,5])
    plt.xlabel('Similarity')
    plt.ylabel('Density')
    plt.hist(sim_vec,density=True)
    plt.savefig(loc,bbox_inches='tight',dpi=700)
    plt.close()


# In[7]:


def add_noise(v,zeta,noise_geometry='hypersphere'):
    # rotate vector v in the direction of a random vector (components drawn from random normal distribution),
    # by a factor drawn from a truncated random normal distribution between 0 and 1, with standard deviation zeta:
    if noise_geometry == 'hypersphere_random_zeta':
        return(np.clip(rotate(v,np.random.normal(size=len(v)),truncnorm(0, 1/zeta, loc=0, scale=zeta).rvs(1)),-1,1))
    if noise_geometry == 'hypersphere':
        return(np.clip(rotate(v,np.random.normal(size=len(v)),zeta),-1,1))
    elif noise_geometry == 'euclidean':
        return(np.clip(v+np.random.normal(size=len(v),scale=zeta),-1,1))


# In[8]:


def initiate_opinion_mat(N=100,S=10,initial_parameters = None):
    """Create a N x S array where the rows are opinion vectors. If initial_parameters == None, return uniform 
    random values from the interval [-1,1]."""
    # If no initial_parameters are given, the opinion space is filled with a uniform distribution:
    if type(initial_parameters) == np.ndarray:
        return(initial_parameters)
    elif initial_parameters == None:
        return(np.random.uniform(low=-1,high=1,size=(N,S)))
    
    # If one initial parameter (float or int) is given, the opinion space is filled with a normal 
    # distribution with mean zero and standard deviation = initial_parameter: 
    elif type(initial_parameters) == float or type(initial_parameters) == int:
        # Draw vector components from random normal distribution:
        m = np.random.normal(size=(N,S))
        # rescale vectors to length from truncated random normal distribution between [0,1]: 
        return(np.apply_along_axis(lambda x: randnormal_rescale(x,0,initial_parameters),1,m))
    
    elif (type(initial_parameters) == list) or (type(initial_parameters) == tuple):
        # If a list or tuple with two parameters is given, the opinion space is filled with vectors
        # whose mean length = initial_parameters[0], and whose length standard deviation is 
        # initial_parameters[1]:
        if len(initial_parameters) == 2:
            # Draw vector components from random normal distribution:
            m = np.random.normal(size=(N,S))
            # rescale vectors to length from truncated random normal distribution between [0,1]: 
            return(np.apply_along_axis(lambda x: randnormal_rescale(x,initial_parameters[0],initial_parameters[1]),1,m))
            
        # if a list or tuple with length one is passed, the same thing is done as if a one parameter 
        # is passed: 
        elif len(initial_parameters) == 1:
            # Draw vector components from random normal distribution:
            m = np.random.normal(size=(N,S))
            # rescale vectors to length from truncated random normal distribution between [0,1]: 
            return(np.apply_along_axis(lambda x: randnormal_rescale(x,0,initial_parameters[0]),1,m))


# In[9]:


def initiate_emo_mat(N=100, initial_parameters = None, min_val=0, max_val=1):
    """Create a length N vector of emotional involvement scores. If initial_parameters == None,
    return uniform random values from the interval [min_val,max_val]. If initial_parameters == [mu,sigma],
    return random values from normal distribution with mean=mu and standard deviation = sigma, 
    truncated between [min_val,max_val]."""
    if (type(initial_parameters) == float) or (type(initial_parameters) == int):
        return(np.array([initial_parameters]*N))
    elif type(initial_parameters) == np.ndarray:
        return(initial_parameters)
    elif type(initial_parameters) == list:
        return(trunc_vec(S=N, sigma=initial_parameters[1], mu=initial_parameters[0], min_val=min_val, max_val=max_val))
    else:
        return(np.random.uniform(low=min_val,high=max_val,size=N))


# In[10]:


def create_matrix(n=12,k=1,score=1,counterscore='reverse',epsilon=0,flip=0,renormalize=True,                  membership=False):
    """Creates n by n Matrix, seperated into k groups that are connected within with strength 'score', and with each 
    other with strength -1*score. If 0<k<1, creates matrix of two groups with relatice size k to (1-k)."""
    # Create Matrix:
    if counterscore == 'reverse':
        counterscore = score*(-1)
    d = (n**2)*[counterscore]
    M = np.array(d)
    M = M.reshape((n,n))
    # group size:

    # Create groups:
    if (k == 1) or (k == 0):
        M[0:,0:] = score
        com_vec = np.array([0] * n)
    elif k > 1:
        k = int(k)
        g_n = int(round(float(n) / k))
        com = 0
        com_vec = np.array([0] * n)
        for i in range(0,n,g_n):
            M[i:(i+g_n),i:(i+g_n)] = score
            com_vec[i:(i+g_n)] = com
            com += 1
    elif 0 < k < 1:
        big_n = int(round(k*n))
        com_vec = np.array([0] * n)
        com_vec[big_n:] = 1 
        M[0:big_n,0:big_n] =  score
        M[big_n:,big_n:] =  score
    elif k == 0:
        M[:,:] = score
    if (flip > 0) and (k != 1):
        x = np.random.choice(n,flip,replace=True)
        y = np.random.choice(n,flip,replace=True)
        for i in range(len(x)):
            if M[x[i],y[i]] == score:
                M[x[i],y[i]] = counterscore
                A,B = np.where(M==counterscore)
                coord = np.random.choice(len(A),1,replace=True)
                a = A[coord]
                b = B[coord]
                M[a,b] = score
            else:
                M[x[i],y[i]] = score
                A,B = np.where(M==score)
                coord = np.random.choice(len(A),1,replace=True)
                a = A[coord]
                b = B[coord]
                M[a,b] = counterscore
    if epsilon > 0:
        M = M + np.random.normal(loc=0.0, scale=epsilon, size=(n,n))
    if (epsilon > 0) and (renormalize == True):
        m = max((np.max(M),abs(np.min(M))))
        M = np.divide(M,m)
    if membership == True:
        return(M,list(com_vec))
    else:
        return(M)


# In[11]:

def average_corr(mat):
    """Average correlation coefficient between columns in mat."""
    # Pearson correlation matrix of transposed input matrix:
    cmat = np.corrcoef(mat.T)
    # Set the diagonal to nan:
    np.fill_diagonal(cmat,np.NaN)
    # Take absolute correlation values:
    cmat = abs(cmat)
    # Fisher Z transformation:
    cmat = np.arctanh(cmat)
    # Average:
    cmean = np.nanmean(cmat)
    # Back-transform to pearson correlation:
    return(np.tanh(cmean))

def mono_pol(mat):
    """Mono-Polarization Index. Computed as average dot product between all unordered pairs of opinion vectors,
    devided by their dimensionality. """
    # Multiply matrix by transposed matrix:
    mulmat = np.matmul(mat,mat.T)
    # Set diagonal to NaN:
    np.fill_diagonal(mulmat,np.NaN)
    # Compute mean, divide by number of dimensions:
    mpol = np.nanmean(mulmat)/mat.shape[1]
    return(mpol)


def polarization(M,distmet='cityblock',ex=2,mink_param=3,max_val=1,min_val=-1,remove_empty_col=True):
    """Computes mean of pairwise distances (of type 'distmet') of row vectors in M without entries representing pair.
    If 'self' is True, also computes distance with self."""
#     if weight_vec:
#         weight_combi_sum = sum([x[0] *x[1] for x in itertools.combinations(weight_vec,2)])
    DistVec = []
    xlist = []
    ylist = []
    pairlist = []
    n = M.shape[0]
    comp_n = M.shape[1]
    ### Determine Maximum and Minimum values, as well as the maximum distance:
    if max_val == 'from_data':
        maval = M.max()
    else:
        maval = max_val
    if min_val == 'from_data':
        mival = M.min()
    else:
        mival = min_val
    DistMax = maval - mival
    
    # If there is no variability in values, return 0 polarization immediatly:
    if DistMax == 0:
        print('error: zero variance in data!')
        return(0)

    # Iterate over all (unordered) pairs of nodes:
    a = range(n)
    #c = 0
    for i in a:
        b = range(i+1,n)
        for j in b:
            #c += 1
            x = M[i,:]
            y = M[j,:]
            # Compute distance between vectors:
            if np.isnan(x).any() or np.isnan(y).any():
                Dist = np.nan
            elif distmet =='cityblock':
                Dist = distance.cityblock(x,y)
            elif distmet == 'minkowski':
                Dist = distance.minkowski(x,y,mink_param)
            elif distmet == 'angle':
                Dist = float(angle_between(x,y))
            elif distmet == 'euclidean':
                Dist = distance.euclidean(x,y)
            elif distmet == 'dot':
                Dist = abs(np.dot(x,np.array(y.T)))
            elif distmet == 'canberra':
                Dist = distance.canberra(x,y)
            elif distmet == 'balance':
                Dist = balance(x.tolist(),y.tolist()).sum()/comp_n
            else:
                print('Error: Unknown Distance Metric')
                break
            DistVec.append(Dist)
    if ex != 1:
        DistVec = [x**ex for x in DistVec]
    max_pairs = (n**2)/4 
    if remove_empty_col == True:
        # Do not count columns that are zero
        comp_n -= np.sum(np.all(np.isnan(M) | np.equal(M, 0), axis=0))
    # Compute normalizing constant for different distance metrics:
    if distmet == 'cityblock':
        normalizer = max_pairs*(comp_n*DistMax)**ex
    elif distmet == 'euclidean':
        normalizer = max_pairs*(sqrt(comp_n*DistMax**2))**(ex)
    elif distmet =='angle':
        if mival < 0:
            normalizer = max_pairs
        # If all values are positive, maximum angle is .25 pi -> adjust normalizer
        else:
            normalizer = max_pairs*((.5)**ex)
    elif distmet =='dot':
        normalizer = max_pairs*(comp_n*(DistMax/2))**ex
    elif distmet == 'balance':
        normalizer = (n*(n-1))/2
    elif distmet == 'canberra':
        normalizer = max_pairs*(comp_n)**ex
    #print(c)
    # Compute sum of squared distances:
    DistSum = np.nansum(DistVec)
    # Normalize sum:
    Pol = float(DistSum)/normalizer
    # return polarization metric:
    # 
    return(Pol)


# In[12]:


def check_threshold(sim,thresh=0):
    if thresh == 0:
        return(True)
    elif type(thresh)  == float:
        return(sim > thresh)
    elif thresh == 'p':
        return(sim > uniform(0,1))
    
        


# In[26]:





# In[43]:

